package de.fhdw.wtf.walker.tasks.test;

import static org.junit.Assert.assertEquals;
import static org.junit.Assert.fail;

import java.util.ArrayList;
import java.util.Collection;

import org.junit.Test;

import de.fhdw.wtf.common.ast.Model;
import de.fhdw.wtf.common.exception.walker.InvalidAttributeModifierException;
import de.fhdw.wtf.common.stream.FilteredTokenStream;
import de.fhdw.wtf.common.stream.SimpleScannerInput;
import de.fhdw.wtf.common.stream.TokenStream;
import de.fhdw.wtf.common.task.TaskExecutorFixed;
import de.fhdw.wtf.common.task.result.ExceptionalTaskResult;
import de.fhdw.wtf.common.task.result.OKTaskResult;
import de.fhdw.wtf.common.task.result.TaskResult;
import de.fhdw.wtf.common.task.result.visitor.TaskResultVisitor;
import de.fhdw.wtf.common.token.Position;
import de.fhdw.wtf.dsl.scanner.common.Scanner;
import de.fhdw.wtf.dsl.scanner.modelScanner.ModelDslScanner;
import de.fhdw.wtf.parser.Parser;
import de.fhdw.wtf.walker.tasks.InvalidAttributeModifierCheck;
import de.fhdw.wtf.walker.tasks.TypeReferencer;
import de.fhdw.wtf.walker.walker.SimpleWalkerTask;

/**
 * Tests the {@link InvalidAttributeModifierCheck} and its exceptions.
 */
public class TestInvalidAttributeModifierCheck {
	
	/**
	 * Group:group=[Class1:class={}; Class2:class={}; Class3:class={Attribute1:Class2 symmetric (Class1);};];.
	 * 
	 * @throws Exception
	 *             : {@link de.fhdw.wtf.common.exception.parser.NoValidTokenStreamException}
	 */
	@Test
	public void testSimpleValidSymmetricModifier() throws Exception {
		final SimpleScannerInput input =
				new SimpleScannerInput("Group:group=[Class1:class={}; Class2:class={};"
						+ "Class3:class={Attribute1:Class2 symmetric (Class1);};];");
		final TokenStream output = FilteredTokenStream.create();
		final Scanner scanner = ModelDslScanner.create();
		scanner.scan(input, output);
		final Parser parser = Parser.create(output);
		final Model model = parser.parse();
		final TaskExecutorFixed taskmanager = TaskExecutorFixed.create();
		final SimpleWalkerTask referencer = TypeReferencer.create(model, taskmanager);
		final InvalidAttributeModifierCheck modifierCheck = InvalidAttributeModifierCheck.create(model, taskmanager);
		modifierCheck.addDependency(referencer);
		taskmanager.startAllKnownTasks();
		
		final Collection<OKTaskResult> okResult = new ArrayList<>();
		final Collection<ExceptionalTaskResult> failResult = new ArrayList<>();
		final Collection<TaskResult> results = taskmanager.getResultsAndShutdown();
		for (final TaskResult current : results) {
			current.accept(new TaskResultVisitor() {
				
				@Override
				public void handleOkTaskResult(final OKTaskResult result) {
					okResult.add(result);
				}
				
				@Override
				public void handleExceptionalTaskResult(final ExceptionalTaskResult result) {
					failResult.add(result);
				}
			});
		}
		
		assertEquals(2, okResult.size());
		assertEquals(0, failResult.size());
	}
	
	/**
	 * Group:group=[Class1:class={}; Class2:class={}; Class3:class={Attribute1:Class2* symmetric (Class1);};];.
	 * 
	 * @throws Exception
	 *             : {@link de.fhdw.wtf.common.exception.parser.NoValidTokenStreamException}
	 */
	@Test
	public void testListOfSymmetricModifier() throws Exception {
		final SimpleScannerInput input =
				new SimpleScannerInput("Group:group=[Class1:class={}; Class2:class={};"
						+ "Class3:class={Attribute1:Class2* symmetric (Class1);};];");
		final TokenStream output = FilteredTokenStream.create();
		final Scanner scanner = ModelDslScanner.create();
		scanner.scan(input, output);
		final Parser parser = Parser.create(output);
		final Model model = parser.parse();
		final TaskExecutorFixed taskmanager = TaskExecutorFixed.create();
		final SimpleWalkerTask referencer = TypeReferencer.create(model, taskmanager);
		final InvalidAttributeModifierCheck modifierCheck = InvalidAttributeModifierCheck.create(model, taskmanager);
		modifierCheck.addDependency(referencer);
		taskmanager.startAllKnownTasks();
		
		final Collection<OKTaskResult> okResult = new ArrayList<>();
		final Collection<ExceptionalTaskResult> failResult = new ArrayList<>();
		final Collection<TaskResult> results = taskmanager.getResultsAndShutdown();
		for (final TaskResult current : results) {
			current.accept(new TaskResultVisitor() {
				
				@Override
				public void handleOkTaskResult(final OKTaskResult result) {
					okResult.add(result);
				}
				
				@Override
				public void handleExceptionalTaskResult(final ExceptionalTaskResult result) {
					failResult.add(result);
				}
			});
		}
		
		assertEquals(2, okResult.size());
		assertEquals(0, failResult.size());
	}
	
	/**
	 * Group:group=[Class1:class={}; Class2:class={}; Class3:class={Attribute1:[Class1->Class2] symmetric (Class4);};
	 * Class4:class={};];.
	 * 
	 * @throws Exception
	 *             : {@link de.fhdw.wtf.common.exception.parser.NoValidTokenStreamException}
	 */
	@Test
	public void testMapOfSymmetricModifier() throws Exception {
		final SimpleScannerInput input =
				new SimpleScannerInput("Group:group=[Class1:class={}; Class2:class={};"
						+ "Class3:class={Attribute1:[Class1->Class2] symmetric (Class4);}; Class4:class={};];");
		final TokenStream output = FilteredTokenStream.create();
		final Scanner scanner = ModelDslScanner.create();
		scanner.scan(input, output);
		final Parser parser = Parser.create(output);
		final Model model = parser.parse();
		final TaskExecutorFixed taskmanager = TaskExecutorFixed.create();
		final SimpleWalkerTask referencer = TypeReferencer.create(model, taskmanager);
		final InvalidAttributeModifierCheck modifierCheck = InvalidAttributeModifierCheck.create(model, taskmanager);
		modifierCheck.addDependency(referencer);
		taskmanager.startAllKnownTasks();
		
		final Collection<OKTaskResult> okResult = new ArrayList<>();
		final Collection<ExceptionalTaskResult> failResult = new ArrayList<>();
		final Collection<TaskResult> results = taskmanager.getResultsAndShutdown();
		for (final TaskResult current : results) {
			current.accept(new TaskResultVisitor() {
				
				@Override
				public void handleOkTaskResult(final OKTaskResult result) {
					okResult.add(result);
				}
				
				@Override
				public void handleExceptionalTaskResult(final ExceptionalTaskResult result) {
					failResult.add(result);
				}
			});
		}
		
		assertEquals(2, okResult.size());
		assertEquals(0, failResult.size());
	}
	
	/**
	 * Group:group=[Class1:class={Attribute:String findable prior transient mutable;};];.
	 * 
	 * @throws Exception
	 *             : {@link de.fhdw.wtf.common.exception.parser.NoValidTokenStreamException}
	 */
	@Test
	public void testModifiersCheck() throws Exception {
		final SimpleScannerInput input =
				new SimpleScannerInput(
						"Group:group=[Class1:class={Attribute:String findable prior transient mutable;};];");
		final TokenStream output = FilteredTokenStream.create();
		final Scanner scanner = ModelDslScanner.create();
		scanner.scan(input, output);
		final Parser parser = Parser.create(output);
		final Model model = parser.parse();
		final TaskExecutorFixed taskmanager = TaskExecutorFixed.create();
		final SimpleWalkerTask referencer = TypeReferencer.create(model, taskmanager);
		final InvalidAttributeModifierCheck modifierCheck = InvalidAttributeModifierCheck.create(model, taskmanager);
		modifierCheck.addDependency(referencer);
		taskmanager.startAllKnownTasks();
		
		final Collection<OKTaskResult> okResult = new ArrayList<>();
		final Collection<ExceptionalTaskResult> failResult = new ArrayList<>();
		final Collection<TaskResult> results = taskmanager.getResultsAndShutdown();
		for (final TaskResult current : results) {
			current.accept(new TaskResultVisitor() {
				
				@Override
				public void handleOkTaskResult(final OKTaskResult result) {
					okResult.add(result);
				}
				
				@Override
				public void handleExceptionalTaskResult(final ExceptionalTaskResult result) {
					failResult.add(result);
				}
			});
		}
		
		assertEquals(2, okResult.size());
		assertEquals(0, failResult.size());
	}
	
	/**
	 * Group:group=[Class1:class={}; Class2:class={}; Class3:class={Attribute1:(class1:Class1, class2:Class2) symmetric
	 * (Class4);}; Class4:class={};];.
	 * 
	 * @throws Exception
	 *             : {@link de.fhdw.wtf.common.exception.parser.NoValidTokenStreamException}
	 */
	@Test
	public void testProductSymmetricModifier() throws Exception {
		final SimpleScannerInput input =
				new SimpleScannerInput("Group:group=[Class1:class={}; Class2:class={}; "
						+ "Class3:class={Attribute1:(class1:Class1, class2:Class2) symmetric (Class4);}; "
						+ "Class4:class={};];");
		final TokenStream output = FilteredTokenStream.create();
		final Scanner scanner = ModelDslScanner.create();
		scanner.scan(input, output);
		final Parser parser = Parser.create(output);
		final Model model = parser.parse();
		final TaskExecutorFixed taskmanager = TaskExecutorFixed.create();
		final SimpleWalkerTask referencer = TypeReferencer.create(model, taskmanager);
		final InvalidAttributeModifierCheck modifierCheck = InvalidAttributeModifierCheck.create(model, taskmanager);
		modifierCheck.addDependency(referencer);
		taskmanager.startAllKnownTasks();
		
		final Collection<OKTaskResult> okResult = new ArrayList<>();
		final Collection<ExceptionalTaskResult> failResult = new ArrayList<>();
		final Collection<TaskResult> results = taskmanager.getResultsAndShutdown();
		for (final TaskResult current : results) {
			current.accept(new TaskResultVisitor() {
				
				@Override
				public void handleOkTaskResult(final OKTaskResult result) {
					okResult.add(result);
				}
				
				@Override
				public void handleExceptionalTaskResult(final ExceptionalTaskResult result) {
					failResult.add(result);
				}
			});
		}
		
		assertEquals(2, okResult.size());
		assertEquals(0, failResult.size());
	}
	
	/**
	 * Group:group=[Class1:class={}; Class2:class={}; Class3:class={Attribute1:{Class1, Class2} symmetric (Class4);};
	 * Class4:class={};];.
	 * 
	 * @throws Exception
	 *             : {@link de.fhdw.wtf.common.exception.parser.NoValidTokenStreamException}
	 */
	@Test
	public void testSumSymmetricModifier() throws Exception {
		final SimpleScannerInput input =
				new SimpleScannerInput("Group:group=[Class1:class={}; Class2:class={}; "
						+ "Class3:class={Attribute1:{Class1, Class2} symmetric (Class4);}; Class4:class={};];");
		final TokenStream output = FilteredTokenStream.create();
		final Scanner scanner = ModelDslScanner.create();
		scanner.scan(input, output);
		final Parser parser = Parser.create(output);
		final Model model = parser.parse();
		final TaskExecutorFixed taskmanager = TaskExecutorFixed.create();
		final SimpleWalkerTask referencer = TypeReferencer.create(model, taskmanager);
		final InvalidAttributeModifierCheck modifierCheck = InvalidAttributeModifierCheck.create(model, taskmanager);
		modifierCheck.addDependency(referencer);
		taskmanager.startAllKnownTasks();
		
		final Collection<OKTaskResult> okResult = new ArrayList<>();
		final Collection<ExceptionalTaskResult> failResult = new ArrayList<>();
		final Collection<TaskResult> results = taskmanager.getResultsAndShutdown();
		for (final TaskResult current : results) {
			current.accept(new TaskResultVisitor() {
				
				@Override
				public void handleOkTaskResult(final OKTaskResult result) {
					okResult.add(result);
				}
				
				@Override
				public void handleExceptionalTaskResult(final ExceptionalTaskResult result) {
					failResult.add(result);
				}
			});
		}
		
		assertEquals(2, okResult.size());
		assertEquals(0, failResult.size());
	}
	
	/**
	 * Group:group=[Class:class={Attribute1:Integer symmetric (String);};];.
	 * 
	 * @throws Exception
	 *             : {@link de.fhdw.wtf.common.exception.parser.NoValidTokenStreamException}
	 */
	@Test
	public void testInvalidSymmetricModifier1() throws Exception {
		final SimpleScannerInput input =
				new SimpleScannerInput("Group:group=[Class:class={Attribute1:Integer symmetric (String);};];");
		final TokenStream output = FilteredTokenStream.create();
		final Scanner scanner = ModelDslScanner.create();
		scanner.scan(input, output);
		final Parser parser = Parser.create(output);
		final Model model = parser.parse();
		final TaskExecutorFixed taskmanager = TaskExecutorFixed.create();
		final SimpleWalkerTask referencer = TypeReferencer.create(model, taskmanager);
		final InvalidAttributeModifierCheck modifierCheck = InvalidAttributeModifierCheck.create(model, taskmanager);
		modifierCheck.addDependency(referencer);
		taskmanager.startAllKnownTasks();
		
		final Collection<OKTaskResult> okResult = new ArrayList<>();
		final Collection<ExceptionalTaskResult> failResult = new ArrayList<>();
		final Collection<TaskResult> results = taskmanager.getResultsAndShutdown();
		for (final TaskResult current : results) {
			current.accept(new TaskResultVisitor() {
				
				@Override
				public void handleOkTaskResult(final OKTaskResult result) {
					okResult.add(result);
				}
				
				@Override
				public void handleExceptionalTaskResult(final ExceptionalTaskResult result) {
					failResult.add(result);
				}
			});
		}
		
		assertEquals(1, okResult.size());
		assertEquals(1, failResult.size());
		
		final Exception exception = failResult.iterator().next().getError();
		if (exception instanceof InvalidAttributeModifierException) {
			final InvalidAttributeModifierException ex = (InvalidAttributeModifierException) exception;
			assertEquals(Position.create("", 1, 46, 45), ex.getStartPosition());
			assertEquals(Position.create("", 1, 63, 62), ex.getEndPos());
		} else {
			fail();
		}
	}
	
}
